/*
 * COPYRIGHT David Koblas, 1990
 * COPYRIGHT Song Li, 1997
 *
 * Originated from David Koblas's GIFTOPPM program.
 *
 */


#include "Copyright.h"
#include "GIFImage.h"
#include "File.h"
#include "Texture.h"





#define GIF_SIG87   "GIF87a"
#define GIF_SIG89   "GIF89a"


#define GIF_SIGLEN       6        // GIF signature length
#define GIF_SCRDESLEN    7        // GIF screen descriptor length
#define GIF_IMGDESLEN    9        // GIF image descriptor length

#define GIF_SEPARATOR   ','       // GIF image separator
#define GIF_EXTENSION   '!'       // GIF extension block marker
#define GIF_TERMINATOR  ';'       // GIF terminator






BBOOL BGIFImage::LoadFileHeader (BFile & file) {

    char sig [GIF_SIGLEN+1];
    if (file.Read (sig, GIF_SIGLEN) != GIF_SIGLEN)
        return BFALSE;
    sig [GIF_SIGLEN] = 0;
    if (! strcmp (sig, GIF_SIG87))
        m_version = GIF87a;
    else
        if (! strcmp (sig, GIF_SIG89))
            m_version = GIF89a;
        else
            return BFALSE;


    BBYTE scrdes [GIF_SCRDESLEN];
    if (file.Read (scrdes, GIF_SCRDESLEN) != GIF_SCRDESLEN)
        return BFALSE;


    m_width     = BREORDER (* (short *) & scrdes [0]);
    m_height    = BREORDER (* (short *) & scrdes [2]);
    m_gpalflag  = (scrdes[4] & 0x80) ? BTRUE : BFALSE;
    m_colorbits = ((scrdes[4] & 0x70) >> 4) + 1;
    m_gpixbits  = (scrdes[4] & 0x07) + 1;
    m_backcolor = scrdes[5];
    m_aspect    = 1.0f;
    m_pixlen    = 1;


    if (m_width < 0)
        m_width = - m_width;
    if (m_height < 0)
        m_height = - m_height;


    if (m_version == GIF87a) {
        if (scrdes[4] & 0x08 || scrdes[6] != 0)
            return BFALSE;
        } 
    else {
        m_gsorted = ((scrdes[4] & 0x08) != 0);
        if (scrdes[6] != 0)
            m_aspect = ((BREAL) scrdes[6] + 15.0f) / 64.0f;
        }


    BASSERT (m_gpalflag);
    BASSERT (m_gpixbits == 8);
    

    m_palsize = 1 << m_gpixbits;
    if (LoadPalette (file) == BFALSE)
        return BFALSE;

    
    return BTRUE;
}









BBOOL BGIFImage::LoadImageHeader (BFile & file) {
    BBYTE separator;
    do {
        if (file.Read (& separator, 1) != 1)
            return BFALSE;
        if (separator == GIF_EXTENSION)
            if (LoadExtension (file) == BFALSE)
                return BFALSE;
    } while (separator == GIF_EXTENSION);


    if (separator != GIF_SEPARATOR)
        return BFALSE;


    BBYTE imgdes [GIF_IMGDESLEN+1];
    if (file.Read (imgdes, GIF_IMGDESLEN) != GIF_IMGDESLEN)
        return BFALSE;


    m_left       = BREORDER (* (short *) & imgdes[0]);
    m_top        = BREORDER (* (short *) & imgdes[2]);
    m_width      = BREORDER (* (short *) & imgdes[4]);
    m_height     = BREORDER (* (short *) & imgdes[6]);
    m_lpalflag   = (imgdes[8] & 0x80) ? BTRUE : BFALSE;
    m_interlaced = (imgdes[8] & 0x40) ? BTRUE : BFALSE;
    m_lpixbits   = (imgdes[8] & 0x07) + 1;


    BASSERT (m_lpalflag == BFALSE);

    if (m_width < 0)
        m_width = - m_width;
    if (m_height < 0)
        m_height = - m_height;


    BBYTE byte;
    if (file.Read (& byte, 1) != 1)
        return BFALSE;

    m_setcodesize = byte;
    m_codesize = m_setcodesize + 1;
    m_clearcode = 1 << m_setcodesize ;
    m_endcode = m_clearcode + 1;
    m_maxcodesize = 2 * m_clearcode;
    m_maxcode = m_clearcode + 2;
    m_curbit = 0;
    m_lastbit = 0;
    m_done = BFALSE;
    m_zerodata = BFALSE;


    int i;
    for (i=0; i<m_clearcode; ++i) {
        m_table[0][i] = 0;
        m_table[1][i] = i;
        }

    for (; i < (1<<12); ++i)
    m_table[0][i] = m_table[1][0] = 0;
    m_stackptr = m_stack;


    return BTRUE;
}






BBOOL BGIFImage::LoadPalette (BFile & file) {
    BASSERT (m_palsize == 256);
    
    BBYTE buffer [3];
    for (int i=0; i<m_palsize; i++) {
        if (file.Read (buffer, 3) != 3)
            return BFALSE;
    
        m_palette [i].r = buffer[0];
        m_palette [i].g = buffer[1];
        m_palette [i].b = buffer[2];
        }

    return BTRUE;
}




BBOOL BGIFImage::LoadExtension (BFile & file) {
    BBYTE byte;
    if (file.Read (& byte, 1) != 1)
        return BFALSE;

    BBYTE count;
    BBYTE buffer [256];
    if (file.Read (& count, 1) != 1)
        return BFALSE;
    while (count > 0) {
        if (file.Read (buffer, count) != count)
            return BFALSE;
        if (file.Read (& count, 1) != 1)
            return BFALSE;
        }
    
    return BTRUE;
}
   






BBOOL BGIFImage::Load (const char * name) {
    BASSERT (m_loaded == BFALSE);

    BFile file;
    if (file.Open (name) == BFALSE)
        return BFALSE;
    if (LoadFileHeader (file) == BFALSE)
        return BFALSE;
    if (LoadImageHeader (file) == BFALSE)
        return BFALSE;


    BASSERT (m_gpalflag);        // Local palette not supported yet.
    BASSERT (! m_interlaced);    // Interlacing not supported yet.
    BASSERT (m_pixlen == 1);     // Pixlen other than 1 not supported yet.


    m_texture = new BTexture (m_width, m_height);
    m_firstpixel = BTRUE;
    for (int y=0; y<m_height; y++)
        for (int x=0; x<m_width; x++) {
            BBYTE pixel;
            if (GetPixel (file, pixel) == BFALSE) {
                delete m_texture;
                m_texture = BNULL;
                file.Close ();
                return BFALSE;
                }
            m_texture->SetPixel (x, m_height-1-y, pixel);
            }


    file.Close ();
    m_loaded = BTRUE;
    return BTRUE;
}




BGIFImage::BGIFImage () :
    m_loaded  (BFALSE),
    m_palette (new BRGB [256]),
    m_texture (BNULL) {

    m_table [0] = new int [1 << 12];
    m_table [1] = new int [1 << 12];
    m_stack     = new int [2 * (1 << 12)];
}
    


BGIFImage::~BGIFImage () {
    delete [] m_stack;
    delete [] m_table [1];
    delete [] m_table [0];
    
    if (m_texture != BNULL)
        delete m_texture;
    delete m_palette;
}





int BGIFImage::GetCode (BFile & file) {
    int	  i, j, ret;
    BBYTE count;

    if (m_curbit + m_codesize >= m_lastbit) {
        if (m_done) {
            if (m_curbit >= m_lastbit)
                return -1;
            }

        m_buffer [0] = m_buffer [m_lastbyte-2];
        m_buffer [1] = m_buffer [m_lastbyte-1];

        if (file.Read (& count, 1) != 1)
            return -1;
        if (count != 0) {
            if (file.Read (m_buffer+2, count) != count)
                return -1;
            }
        else {
            m_zerodata = BTRUE;
            m_done = BTRUE;
            }

        m_lastbyte = 2 + count;
        m_curbit = (m_curbit - m_lastbit) + 16;
        m_lastbit = (2 + count) * 8;
        }


    ret = 0;
    for (i=m_curbit, j=0; j<m_codesize; ++i, ++j)
        ret |= ((m_buffer[i/8] & (1 << (i%8))) != 0) << j;
    m_curbit += m_codesize;
    return ret;
}





BBOOL BGIFImage::GetPixel (BFile & file, BBYTE & pix) {
    int code, incode;

    if (m_firstpixel) {
        m_firstpixel = BFALSE;
	do {
            m_firstcode = m_prevcode = GetCode(file);
        } while (m_firstcode == m_clearcode);
        return m_firstcode;
        }


    if (m_stackptr > m_stack) {
        pix = * --m_stackptr;
        return BTRUE;
        }

    while ((code = GetCode (file)) >= 0) {
        if (code == m_clearcode) {
            int i;
            for (i=0; i<m_clearcode; ++i) {
                m_table[0][i] = 0;
                m_table[1][i] = i;
                }

            for (; i < (1<<12); ++i)
                m_table[0][i] = m_table[1][i] = 0;
            m_codesize = m_setcodesize + 1;
            m_maxcodesize = 2 * m_clearcode;
            m_maxcode = m_clearcode + 2;
            m_stackptr = m_stack;
            m_firstcode = m_prevcode = GetCode (file);
            pix = m_firstcode;
            return BTRUE;
            }
        else
            if (code == m_endcode) {
                if (m_zerodata)
                    return BFALSE;
                
                BBYTE buffer [260];
                BBYTE datacount;

                do {
                    if (file.Read (& datacount, 1) != 1)
                        return BFALSE;
                    if (datacount != 0)
                        if (file.Read (buffer, datacount) != datacount)
                            return BFALSE;
                } while (datacount > 0);
            }


        incode = code;


        if (code >= m_maxcode) {
            *m_stackptr++ = m_firstcode;
            code = m_prevcode;
            }

        while (code >= m_clearcode) {
            *m_stackptr++ = m_table[1][code];
            code = m_table[0][code];
            }


        *m_stackptr++ = m_firstcode = m_table[1][code];
        if ((code = m_maxcode) < (1<<12)) {
            m_table[0][code] = m_prevcode;
            m_table[1][code] = m_firstcode;
            ++ m_maxcode;
            if ((m_maxcode >= m_maxcodesize) &&
                (m_maxcodesize < (1<<12))) {
                m_maxcodesize *= 2;
                ++ m_codesize;
                }
            }

        m_prevcode = incode;
        if (m_stackptr > m_stack) {
            pix = * --m_stackptr;
            return BTRUE;
            }
        }


    pix = code;
    return BTRUE;
}




