#include "Copyright.h"
#include "Rengine.h"
#include "Vector.h"
#include "Point.h"
#include "Texture.h"
#include "Buffer.h"
#include "Camera.h"
#include "Line.h"
#include "Sector.h"



#define  XTEXTRATIO     12.8f
#define  YTEXTRATIO     12.8f



static const BCamera *  curcamera;
static unsigned int     curtime = 0;
static BBuffer *        curbuffer;
static int              curcol;
static BVector          curray;


static BREAL curpaintvx;
static BREAL curpaintvy;
static BREAL curpaintpx;
static BREAL curpaintpy;





BRengine::BRengine (int width, int height, 
                    BREAL distance, BREAL projector) :
    m_width      (width),
    m_height     (height),
    m_zoomer     (distance * projector),
    m_projector  (projector),
    m_raybase    (new BVector [width])  {
               
    BASSERT (width % 4 == 0);

    int i;
    for (i=0; i<width; i++) {
        m_raybase[i].x = distance;
        m_raybase[i].y = (BREAL) (width/2 - i);
        }
}


                      

BRengine::~BRengine () {
    delete [] m_raybase;
}







void BRengine::PaintWall (const BTexture * texture,
                          BREAL minlevel,  BREAL maxlevel,
                          BREAL xshift,    BREAL yshift,
                          BREAL perspector,
                          BREAL alphalen) const {

    BASSERT (texture != BNULL);
    
    const int   textcol  = (int) (XTEXTRATIO * (xshift + alphalen));
    const BREAL textmin  = YTEXTRATIO * 
                           (curcamera->m_vlevel + yshift + minlevel/perspector);
    const int   bufmin   = m_height/2 + (int) minlevel;
    const int   bufmax   = m_height/2 + (int) maxlevel;
    const BREAL factor   = YTEXTRATIO / perspector;
    
    
    BASSERT (bufmin  >= 0 && bufmin  < m_height);
    BASSERT (bufmax  >= 0 && bufmax  < m_height);
    BASSERT (textcol >= 0);
    BASSERT (bufmin  <= bufmax);



    int row;
    int height = bufmax-bufmin;
    BREAL textrow = textmin + height * factor;
    for (row = bufmax;
         height >= 0;
         row--, height--, textrow -= factor)
         curbuffer->SetPixel (curcol, row, 
                              texture->GetPixel (textcol, (int) textrow));
}





void BRengine::PaintSector (const BTexture * texture, 
                            BREAL minrow, BREAL maxrow,
                            BREAL xshift, BREAL yshift,
                            BREAL level) const {

    const int bufmin = (int) minrow;
    const int bufmax = (int) maxrow;
    const BREAL paintpx = curpaintpx + XTEXTRATIO * xshift;
    const BREAL paintpy = curpaintpy + YTEXTRATIO * yshift;
    const BREAL paintvx = level * curpaintvx;
    const BREAL paintvy = level * curpaintvy;
    
    BASSERT (bufmin >= -m_height/2 && bufmin <= m_height/2);
    BASSERT (bufmax >= -m_height/2 && bufmax <= m_height/2);
    BASSERT ((0.0f <= bufmin && bufmin <= bufmax) ||
             (0.0f >= bufmax && bufmax >= bufmin));
    
    for (register int row=bufmin; row<=bufmax; row++)
        curbuffer->SetPixel (curcol, m_height/2+row,
            texture->GetPixel ((int) (paintpx + paintvx/row),
                               (int) (paintpy + paintvy/row)));
}


      
    

void BRengine::SubRender (BREAL minrow, BREAL maxrow, 
                          const BPoint  & startpoint,
                          const BSector * sector, 
                          BREAL basebeta) const {

    BASSERT (minrow <= maxrow);
    BASSERT (sector != BNULL);

   /*
    * STEP 1:  OUTPUTS - line, beta, alpha.
    * 
    */ 


    int   i;
    BREAL beta  = BLIMMAX;
    BREAL alpha;
    BLine * line = BNULL;
    BFACE face;


    for (i=0; i<sector->m_linecount; i++) {
        BLine * curline = sector->Line (i);
        if (curline->m_timestamp < curtime) {
            BREAL curalpha;
            BREAL curbeta;
            BFACE curface  = sector->Face (i);
            BREAL iproduct = curray.y * curline->m_vector.x - 
                             curray.x * curline->m_vector.y ;

            if ((curface == BFRONT && iproduct <= 0.0f) ||
                (curface == BBACK  && iproduct >= 0.0f))
                continue;


            curline->CrossTest (startpoint, curray, 
                                iproduct, curalpha, curbeta);


                
            if (curbeta >= 0.0f && curalpha >= 0.0f && curalpha <= 1.0f)
                if (beta >= curbeta) {
                    if (beta == curbeta) {
                        BASSERT (BNEAR(alpha,0.0f) || BNEAR(alpha,1.0f));
                        BASSERT (BNEAR(curalpha,0.0f) || BNEAR(curalpha,1.0f));
                        BVector curvec = BNEAR (curalpha, 0.0f) ? 
                                         curline->m_vector : - curline->m_vector;
                        BVector vec    = BNEAR (alpha, 0.0f) ? 
                                         line->m_vector : - line->m_vector;
                        if (curvec * curray / curvec.Length () > 
                            vec    * curray / vec.Length ())
                            continue;
                        }
                            
                    face  = curface;
                    line  = curline;
                    beta  = curbeta;
                    alpha = curalpha;
                    }
            }
        }




    BASSERT (line != BNULL);
    BASSERT (alpha >= 0.0f && alpha <= 1.0f);
    BASSERT (beta >= 0.0f);

    line->m_timestamp = curtime;



   /*
    * STEP 2:  OUTPUTS - top, bottom, ceil, floor.
    *
    */ 

    const BSector * nextsec = line->m_face[BOTHER (face)].m_sector;


    BREAL top    = sector->m_ceillevel;
    BREAL bottom = sector->m_floorlevel;
    if (nextsec != BNULL) {
        if (nextsec->m_ceillevel < sector->m_ceillevel)
            top = BMAX (nextsec->m_ceillevel, sector->m_floorlevel);
        if (nextsec->m_floorlevel > sector->m_floorlevel)
            bottom = BMIN (nextsec->m_floorlevel, sector->m_ceillevel);
        }


    BREAL ceil  = sector->m_ceillevel  - curcamera->m_vlevel;
    BREAL floor = sector->m_floorlevel - curcamera->m_vlevel;
    top    -= curcamera->m_vlevel;
    bottom -= curcamera->m_vlevel;

    BASSERT (floor  <= bottom);
    BASSERT (bottom <= top);
    BASSERT (top    <= ceil);



   /*
    * STEP 3: Render recursively.
    *
    */ 


    BREAL alphalen   = alpha * line->Length ();
    BREAL perspector = m_zoomer / BMAX(beta+basebeta,BEPSILON);
    BREAL scnceil    = ceil   * perspector;
    BREAL scnfloor   = floor  * perspector;
    BREAL scntop     = top    * perspector;
    BREAL scnbottom  = bottom * perspector;
    

    BASSERT (scnbottom <= scntop);
    BASSERT (scnbottom >= scnfloor);
    BASSERT (scntop    <= scntop);


    if (scnceil > maxrow)  scnceil = maxrow;
        else if (scnceil < minrow)  scnceil = minrow;
    if (scnfloor > maxrow)  scnfloor = maxrow; 
        else if (scnfloor<minrow) scnfloor = minrow;
    if (scntop > maxrow)  scntop = maxrow;
        else if (scntop < minrow)  scntop = minrow;
    if (scnbottom > maxrow)  scnbottom = maxrow;
        else if (scnbottom < minrow)  scnbottom = minrow;



    BASSERT (scnceil   <= maxrow);
    BASSERT (scnfloor  >= minrow);
    BASSERT (scnbottom <= scntop);
    BASSERT (scnbottom >= scnfloor);
    BASSERT (scntop    <= scntop);


    if (scnceil < maxrow)
        PaintSector (sector->m_ceiltext, 
                     scnceil, maxrow, 
                     sector->m_cxshift, sector->m_cyshift,
                     ceil);

    if (scnfloor > minrow)
        PaintSector (sector->m_floortext, 
                     minrow, scnfloor, 
                     sector->m_fxshift, sector->m_fyshift,
                     floor);
 

    if (scnceil != scntop)
        PaintWall (line->m_face[face].m_utext, 
                   scntop, scnceil, 
                   line->m_face[face].m_xshift, 
                   line->m_face[face].m_yshift,
                   perspector, alphalen);

    
    if (scnfloor != scnbottom)
        PaintWall (line->m_face[face].m_ltext, 
                   scnfloor, scnbottom,
                   line->m_face[face].m_xshift, 
                   line->m_face[face].m_yshift,
                   perspector, alphalen);


    if (scnbottom != scntop) {
        const BTexture * texture = line->m_face[face].m_mtext;
        if (texture == BNULL)
            SubRender (scnbottom, scntop, 
                       *line->m_start + line->m_vector * alpha,
                       nextsec, beta+basebeta);
        else
            PaintWall (texture, scnbottom, scntop,
                       line->m_face[face].m_xshift, 
                       line->m_face[face].m_yshift,
                       perspector, alphalen);
        }
}





void BRengine::Render (BBuffer & buffer, const BCamera & camera) const {

    BASSERT (m_width  == buffer.Width  ());
    BASSERT (m_height == buffer.Height ());

    curbuffer = & buffer;
    curcamera = & camera;
    curpaintpx = XTEXTRATIO * camera.m_vpoint.x;
    curpaintpy = YTEXTRATIO * camera.m_vpoint.y;
    BREAL sintheta = camera.m_vnormal.y / camera.m_vnormal.Length ();
    BREAL costheta = camera.m_vnormal.x / camera.m_vnormal.Length ();


    for (int col=0; col<m_width; col++) {
        curtime ++;
        curcol = col;
        curray.x = costheta * m_raybase[col].x - sintheta * m_raybase[col].y;
        curray.y = sintheta * m_raybase[col].x + costheta * m_raybase[col].y;
        curpaintvx = curray.x * (XTEXTRATIO * m_zoomer);
        curpaintvy = curray.y * (YTEXTRATIO * m_zoomer);

        SubRender (-(BREAL)m_height/2.0f, (BREAL)m_height/2.0f-1.0f, 
                   camera.m_vpoint,
                   camera.m_vsector, 0.0f);
        }

}

